#cs ----------------------------------------------------------------------------

 AutoIt Version: 3.3.16.1
 Author:         Kanashius

 Script Function:
	An UDF to generate and draw QRCodes.
	This UDF is a wrapper for the library provided by https://www.nayuki.io/page/qr-code-generator-library.
	GitHub: https://github.com/nayuki/QR-Code-generator

#ce ----------------------------------------------------------------------------
#include-once
#include <GDIPlus.au3>

; ECL = error correction level
Global $_QRCode_ECL_LOW = 1
Global $_QRCode_ECL_MEDIUM = 2
Global $_QRCode_ECL_QUARTILE = 3
Global $_QRCode_ECL_HIGH = 4

Global $_QRCode_TagQRCode = "struct;byte error;uint width;uint height;ptr data;uint neededbits;uint possiblebits;endstruct"
Global $_QRCode_DLL = -1, $_QRCode_tQRCode = -1, $_QRCode_DLL_Folder = "", $_QRCode_DLL_FileName="qrcode-x32.dll"
If @AutoItX64 Then $_QRCode_DLL_FileName="qrcode-x64.dll"


; #FUNCTION# =======================================================================================
; Name ..........: _QRCode_StartUp
; Description ...: Start up of the QRCode UDF. This includes calling _GDIPlus_Startup().
; Syntax ........: _QRCode_StartUp($sFolder = $_QRCode_DLL_Folder, $sFile = $_QRCode_DLL_FileName)
; Parameters ....: $sFolder               - The folder where the dll can be found (with \ terminated, if not empty). Default: ""
;                  $sFile                 - The filename of the dll. Default: "qrcode-x32.dll" or "qrcode-x64.dll", depending on the executable type (@AutoItX64)
; Return values .: True on success. False when an error occurred.
; Errors ........: 1 - If the dll failed to load. @extended contains the error code from DllOpen().
;                  2 - If the QRCode struct failed to create. @extended contains the error code from DllStructCreate().
; Author ........: Kanashius
; ==================================================================================================
Func _QRCode_StartUp($sFolder = $_QRCode_DLL_Folder, $sFile = $_QRCode_DLL_FileName)
	_GDIPlus_Startup()
	If $_QRCode_DLL <> -1 Then DllClose($_QRCode_DLL)
	$_QRCode_DLL = DllOpen($sFolder&$sFile)
	If @error or $_QRCode_DLL=-1 Then return SetError(1, @error, False)
	$_QRCode_tQRCode = DllStructCreate($_QRCode_TagQRCode)
	If @error Then return SetError(2, @error, False)
	return True
EndFunc

; #FUNCTION# =======================================================================================
; Name ..........: _QRCode_Shutdown
; Description ...: Shutdown the QRCode UDF. This includes calling _GDIPlus_Shutdown().
; Syntax ........: _QRCode_Shutdown()
; Return values .: True on success. False when an error occurred.
; Errors ........: 1 - Destroying the QRCode struct failed.
; Author ........: Kanashius
; ==================================================================================================
Func _QRCode_Shutdown()
	_GDIPlus_Shutdown()
	$ptrResult = DllCall($_QRCode_DLL, "none:cdecl", "Clean", "ptr", DllStructGetPtr($_QRCode_tQRCode))
	Local $iError = @error
	DllClose($_QRCode_DLL)
	$_QRCode_DLL = -1
	If $iError Then return SetError(1, $iError, True)
	return False
EndFunc

; #FUNCTION# =======================================================================================
; Name ..........: _QRCode_GetBitmap
; Description ...: Creates a GDIPlus-Bitmap with the QRCode.
; Syntax ........: _QRCode_GetBitmap(ByRef $arQrCode, $iPixelSize = 1, $iBorderTop = 0, $iBorderRight = 0, $iBorderBottom = 0, $iBorderLeft = 0, $iColor = 0xFF000000, $iColorBG = 0xFFFFFFFF)
; Parameters ....: $arQRCode              - The QRCode array (2D-Array with bool values)
;                  $iPixelSize            - The size of each rectangle to be drawn
;                  $iBorderTop            - The space in pixels to the top
;                  $iBorderRight          - The space in pixels to the right
;                  $iBorderBottom         - The space in pixels to the bottom
;                  $iBorderLeft           - The space in pixels to the left
;                  $iColor                - The ARGB color of the foreground (normally black)
;                  $iColorBG              - The ARGB color of the background (normally white)
; Return values .: The created Bitmap. False when an error occurred
; Errors ........: 1 - The input parameters were wrong. @extended contains the number of the parameter, which was wrong (starting with 1).
;                  2 - Error calling a _gdiplus_* method. @extended provides more details.
;                  3 - Error calling _QRCode_DrawQRCode. @extended contains the error code thrown by _QRCode_DrawQRCode.
; Author ........: Kanashius
; ==================================================================================================
Func _QRCode_GetBitmap(ByRef $arQrCode, $iPixelSize = 1, $iBorderTop = 0, $iBorderRight = 0, $iBorderBottom = 0, $iBorderLeft = 0, $iColor = 0xFF000000, $iColorBG = 0xFFFFFFFF)
	If Not IsArray($arQrCode) or UBound($arQrCode, 0)<>2 Then return SetError(1, 2, False)
	If Not IsNumber($iPixelSize) Then return SetError(1, 3, False)
	If Not IsNumber($iBorderTop) Then return SetError(1, 4, False)
	If Not IsNumber($iBorderRight) Then return SetError(1, 5, False)
	If Not IsNumber($iBorderBottom) Then return SetError(1, 6, False)
	If Not IsNumber($iBorderLeft) Then return SetError(1, 7, False)
	If Not IsNumber($iColor) Then return SetError(1, 8, False)
	If Not IsNumber($iColorBG) Then return SetError(1, 9, False)
	Local $iWidth = $iBorderLeft+$iBorderRight+UBound($arQrCode, 2)*$iPixelSize
	Local $iHeight = $iBorderTop+$iBorderBottom+UBound($arQrCode)*$iPixelSize
	Local $hBitmap = _GDIPlus_BitmapCreateFromScan0($iWidth, $iHeight)
	If @error Then return SetError(2, 1, False)
	Local $hContext = _GDIPlus_ImageGetGraphicsContext($hBitmap)
	If @error Then return SetError(2, 2, False)
	_GDIPlus_GraphicsClear($hContext, 0x00FFFFFF)
	If @error Then return SetError(2, 3, False)
	_QRCode_DrawQRCode($hContext, $arQrCode, $iPixelSize, $iBorderLeft, $iBorderTop, $iColor, $iColorBG)
	If @error Then return SetError(3, @error, -1)
	_GDIPlus_GraphicsDispose($hContext)
	If @error Then return SetError(2, 4, False)
	return $hBitmap
EndFunc

; #FUNCTION# =======================================================================================
; Name ..........: _QRCode_DrawQRCode
; Description ...: Draws a QRCode on the provided GraphicsContext
; Syntax ........: _QRCode_DrawQRCode(ByRef $hGraphicsContext, ByRef $arQRCode, $iPixelSize = 1, $iOffsetLeft = 0, $iOffestTop = 0, $iColor = 0xFF000000, $iColorBG = 0xFFFFFFFF)
; Parameters ....: $hGraphicsContext      - The graphics context to draw on
;                  $arQRCode              - The QRCode array (2D-Array with bool values)
;                  $iPixelSize            - The size of each rectangle to be drawn
;                  $iOffsetLeft           - The space in pixels to the left, before the qrcode is drawn
;                  $iOffestTop            - The space in pixels to the top, before the qrcode is drawn
;                  $iColor                - The ARGB color of the foreground (normally black)
;                  $iColorBG              - The ARGB color of the background (normally white)
; Return values .: True on success, False when an error occurred
; Errors ........: 1 - The input parameters were wrong. @extended contains the number of the parameter, which was wrong (starting with 1).
;                  2 - Error calling a _gdiplus_* method. @extended provides more details.
;                  3 - Error calling _GDIPlus_GraphicsFillRect to draw a rectangle (@extended: 1 - foreground, 2 - background)
; Author ........: Kanashius
; ==================================================================================================
Func _QRCode_DrawQRCode(ByRef $hGraphicsContext, ByRef $arQRCode, $iPixelSize = 1, $iOffsetLeft = 0, $iOffestTop = 0, $iColor = 0xFF000000, $iColorBG = 0xFFFFFFFF)
	If Not IsPtr($hGraphicsContext) Then return SetError(1, 1, False)
	If Not IsArray($arQrCode) or UBound($arQrCode, 0)<>2 Then return SetError(1, 2, False)
	If Not IsNumber($iPixelSize) Then return SetError(1, 3, False)
	If Not IsNumber($iOffsetLeft) Then return SetError(1, 4, False)
	If Not IsNumber($iOffestTop) Then return SetError(1, 5, False)
	If Not IsNumber($iColor) Then return SetError(1, 6, False)
	If Not IsNumber($iColorBG) Then return SetError(1, 7, False)
	Local $brushColor = _GDIPlus_BrushCreateSolid(0xFF000000)
	If @error Then return SetError(2, 1, False)
	Local $brushBG = _GDIPlus_BrushCreateSolid(0xFFFFFFFF)
	If @error Then return SetError(2, 2, False)
	Local $iSmoothingMode = _GDIPlus_GraphicsGetSmoothingMode($hGraphicsContext)
	If @error Then return SetError(2, 3, False)
	_GDIPlus_GraphicsSetSmoothingMode($hGraphicsContext, 0)
	If @error Then return SetError(2, 4, False)
	For $x=0 to UBound($arQRCode)-1 Step 1
		For $y=0 to UBound($arQRCode, 2)-1 Step 1
			If $arQRCode[$x][$y] Then
				_GDIPlus_GraphicsFillRect($hGraphicsContext, $iOffsetLeft+$x*$iPixelSize, $iOffestTop+$y*$iPixelSize, $iPixelSize, $iPixelSize, $brushColor)
				If @error Then return SetError(3, 1, False)
			Else
				_GDIPlus_GraphicsFillRect($hGraphicsContext, $iOffsetLeft+$x*$iPixelSize, $iOffestTop+$y*$iPixelSize, $iPixelSize, $iPixelSize, $brushBG)
				If @error Then return SetError(3, 2, False)
			EndIf
		Next
	Next
	_GDIPlus_GraphicsSetSmoothingMode($hGraphicsContext, $iSmoothingMode)
	If @error Then return SetError(2, 5, False)
	_GDIPlus_BrushDispose($brushColor)
	If @error Then return SetError(2, 6, False)
	_GDIPlus_BrushDispose($brushBG)
	If @error Then return SetError(2, 7, False)
	return True
EndFunc

; #FUNCTION# =======================================================================================
; Name ..........: _QRCode_DrawQRCodeFast
; Description ...: Draws a QRCode on the provided GraphicsContext.
;                  WARNING: This fast function requires the _GDIPlus_Brushes to be provided and does not check any parameters or function calls for errors.
; Syntax ........: _QRCode_DrawQRCodeFast(ByRef $hGraphicsContext, ByRef $arQRCode, ByRef $hColorBrush, ByRef $hColorBGBrush, $iPixelSize = 1, $iOffsetLeft = 0, $iOffestTop = 0)
; Parameters ....: $hGraphicsContext      - The graphics context to draw on
;                  $arQRCode              - The QRCode array (2D-Array with bool values)
;                  $hColorBrush           - The GDIPlus Brush of the foreground (normally black)
;                  $hColorBGBrush         - The GDIPlus Brush of the background (normally white)
;                  $iPixelSize            - The size of each rectangle to be drawn
;                  $iOffsetLeft           - The space in pixels to the left, before the qrcode is drawn
;                  $iOffestTop            - The space in pixels to the top, before the qrcode is drawn
; Return values .: True
; Author ........: Kanashius
; ==================================================================================================
Func _QRCode_DrawQRCodeFast(ByRef $hGraphicsContext, ByRef $arQRCode, ByRef $hColorBrush, ByRef $hColorBGBrush, $iPixelSize = 1, $iOffsetLeft = 0, $iOffestTop = 0)
	Local $iSmoothingMode = _GDIPlus_GraphicsGetSmoothingMode($hGraphicsContext)
	_GDIPlus_GraphicsSetSmoothingMode($hGraphicsContext, 0)
	For $x=0 to UBound($arQRCode)-1 Step 1
		For $y=0 to UBound($arQRCode, 2)-1 Step 1
			If $arQRCode[$x][$y] Then
				_GDIPlus_GraphicsFillRect($hGraphicsContext, $iOffsetLeft+$x*$iPixelSize, $iOffestTop+$y*$iPixelSize, $iPixelSize, $iPixelSize, $hColorBrush)
			Else
				_GDIPlus_GraphicsFillRect($hGraphicsContext, $iOffsetLeft+$x*$iPixelSize, $iOffestTop+$y*$iPixelSize, $iPixelSize, $iPixelSize, $hColorBGBrush)
			EndIf
		Next
	Next
	_GDIPlus_GraphicsSetSmoothingMode($hGraphicsContext, $iSmoothingMode)
	return True
EndFunc

; #FUNCTION# =======================================================================================
; Name ..........: _QRCode_ConsoleWrite
; Description ...: Prints the QRCode to the console.
; Syntax ........: _QRCode_GetQRCode(ByRef $arQrCode, $sColorChar = "#", $sColorBG = " ", $sLineBreak = @CRLF)
; Parameters ....: $arQrCode               - The QRCode array (2D-Array with bool values)
;                  $sColorChar             - The character for the foreground color (normally black)
;                  $sColorBG               - The character for the background color (normally white)
;                  $sLineBreak             - The character for line breaks
;                  $iSize                  - Number of times the $sColorChar or $sColorBG are repeated. Default is 2 because width is smaller then space => more square
;                  $iBorderTop             - Number of empty lines before the output
;                  $iBorderRight           - Number of spaces before the output
;                  $iBorderBottom          - Number of spaces after the output
;                  $iBorderLeft            - Number of empty lines after the output
; Return values .: True on success, False when an error occurred
; Errors ........: 1 - The input parameters were wrong. @extended contains the number of the parameter, which was wrong (starting with 1).
;                      Extended: 1 - The provided QRCode array is not an Array or not a 2D-Array.
;                      Extended: 2 - The provided size is not an integer >0
;                      Extended: 3 - The provided $iBorderTop is not an integer >=0
;                      Extended: 4 - The provided $iBorderRight is not an integer >=0
;                      Extended: 5 - The provided $iBorderBottom is not an integer >=0
;                      Extended: 6 - The provided $iBorderLeft is not an integer >=0
; Author ........: Kanashius
; ==================================================================================================
Func _QRCode_ConsoleWrite(ByRef $arQrCode, $sColorChar = "#", $sColorBG = " ", $sLineBreak = @CRLF, $iSize = 2, $iBorderTop = 0, $iBorderRight = 0, $iBorderBottom = 0, $iBorderLeft = 0)
	If Not IsArray($arQrCode) or UBound($arQrCode, 0)<>2 Then return SetError(1, 1, False)
	If Not IsInt($iSize) or $iSize<1 Then return SetError(1, 2, False)
	If Not IsInt($iBorderTop) or $iBorderTop<0 Then return SetError(1, 3, False)
	If Not IsInt($iBorderRight) or $iBorderRight<0 Then return SetError(1, 4, False)
	If Not IsInt($iBorderBottom) or $iBorderBottom<0 Then return SetError(1, 5, False)
	If Not IsInt($iBorderLeft) or $iBorderLeft<0 Then return SetError(1, 6, False)
	For $i=0 to $iBorderTop-1 Step 1
		ConsoleWrite(@crlf)
	Next
	For $x=0 to UBound($arQRCode)-1 Step 1
		For $i=0 to $iBorderLeft-1 Step 1
			ConsoleWrite(" ")
		Next
		For $y=0 to UBound($arQRCode, 2)-1 Step 1
			If $arQRCode[$x][$y] Then
				For $i=0 to $iSize-1 Step 1
					ConsoleWrite($sColorChar)
				Next
			Else
				For $i=0 to $iSize-1 Step 1
					ConsoleWrite($sColorBG)
				Next
			EndIf
		Next
		For $i=0 to $iBorderRight-1 Step 1
			ConsoleWrite(" ")
		Next
		ConsoleWrite($sLineBreak)
	Next
	For $i=0 to $iBorderBottom-1 Step 1
		ConsoleWrite(@crlf)
	Next
	return True
EndFunc

; #FUNCTION# =======================================================================================
; Name ..........: _QRCode_GetSize
; Description ...: Get the size of the qrcode.
; Syntax ........: _QRCode_GetSize(ByRef $arQrCode, $iPixelSize = 1)
; Parameters ....: $arQrCode               - The QRCode array (2D-Array with bool values)
;                  $iPixelSize             - The size of each rectangle to be drawn in pixels => multiplier for the size
; Return values .: The Array with the sizes: [width, height]. False when an error occurred
; Errors ........: 1 - The input parameters were wrong. @extended contains the number of the parameter, which was wrong (starting with 1).
; Author ........: Kanashius
; ==================================================================================================
Func _QRCode_GetSize(ByRef $arQrCode, $iPixelSize = 1)
	If Not IsArray($arQrCode) or UBound($arQrCode, 0)<>2 Then return SetError(1, 2, False)
	If Not IsNumber($iPixelSize) Then return SetError(1, 3, False)
	Local $arSize[2]
	$arSize[0] = UBound($arQrCode, 2)*$iPixelSize
	$arSize[1] = UBound($arQrCode)*$iPixelSize
	return $arSize
EndFunc

; #FUNCTION# =======================================================================================
; Name ..........: _QRCode_GetECLMaxLoss
; Description ...: Returns the maximum allowed loss before the qrcode becomes unreadable as decimal multiplier.
; Syntax ........: _QRCode_GetECLMaxLoss($iErrorLevel=$_QRCode_ECL_HIGH)
; Parameters ....: $iErrorLevel           - The error correction level (ECL). The higher, the more errors can happen, without causing problems,
;                                           but a higher ECL also means less data fitting inside the QRCode and the QRCode gets larger.
;                                           Order (low to high) with amount of allowed loss in percent:
;                                           $_QRCode_ECL_LOW (~7% => 0.07), $_QRCode_ECL_MEDIUM (~15% => 0.15), $_QRCode_ECL_QUARTILE (~25% => 0.25), $_QRCode_ECL_HIGH (~30% => 0.3)
;                                           Example: A higher ECL allows for adding larger icons without loosing the data.
; Return values .: The percentage of allowed loss as decimal multiplicator (1% => 0.01)
; Errors ........: 1 - The input parameters were wrong.
;                      Extended: 1 - The Error level is not one of the required values: $_QRCode_ECL_LOW/$_QRCode_ECL_MEDIUM/$_QRCode_ECL_QUARTILE/$_QRCode_ECL_HIGH
; Author ........: Kanashius
; ==================================================================================================
Func _QRCode_GetECLMaxLoss($iErrorLevel=$_QRCode_ECL_HIGH)
	If Not ($iErrorLevel=$_QRCode_ECL_HIGH or $iErrorLevel=$_QRCode_ECL_LOW or $iErrorLevel=$_QRCode_ECL_MEDIUM or $iErrorLevel=$_QRCode_ECL_QUARTILE) Then return SetError(1, 1, -1)
	Switch $iErrorLevel
		Case $_QRCode_ECL_LOW
			return 0.07
		Case $_QRCode_ECL_MEDIUM
			return 0.15
		Case $_QRCode_ECL_QUARTILE
			return 0.25
		Case $_QRCode_ECL_HIGH
			return 0.3
	EndSwitch
EndFunc

; #FUNCTION# =======================================================================================
; Name ..........: _QRCode_GetQRCode
; Description ...: Generates a QRCode as 2D-Boolean-Array, where the first index represents the rows (lines) and the second index represents the columns.
;                  False represents the Background (normally white) and True represents the Foreground (normally black).
; Syntax ........: _QRCode_GetQRCode($sText, $sErrorLvl = $_QRCode_ECL_HIGH)
; Parameters ....: $sText                 - The text to put in the qrcode
;                  $iErrorLevel           - The error correction level (ECL). The higher, the more errors can happen, without causing problems,
;                                           but a higher ECL also means less data fitting inside the QRCode and the QRCode gets larger.
;                                           Order (low to high) with amount of allowed loss in percent:
;                                           $_QRCode_ECL_LOW (~7%), $_QRCode_ECL_MEDIUM (~15%), $_QRCode_ECL_QUARTILE (~25%), $_QRCode_ECL_HIGH (~30%)
;                                           Example: A higher ECL allows for adding larger icons without loosing the data.
; Return values .: The QRCode as 2D-Array with bool values; Returns -1 if an error occurred
; Errors ........: 1 - The input parameters were wrong.
;                      Extended: 1 - The Text is not a String
;                      Extended: 2 - The Error level is not one of the required values: $_QRCode_ECL_LOW/$_QRCode_ECL_MEDIUM/$_QRCode_ECL_QUARTILE/$_QRCode_ECL_HIGH
;                  2 - Call to qrcode-??.dll failed. @extended contains the DllCall error.
;                  3 - Creation of the retrieval boolean[?] struct failed. @extended contains the DllStructCreate error.
;                  4 - The qrcode dll failed to read $sText.
;                  5 - There was a segment in the qrcode, which was too long
;                  6 - The given data was to much. @extended contains the difference between the needed bit and the possible bits.
;                  7 - An unknown error in the qrcode dll occurred. (Can not happen, until a new version is released)
;                  8 - The QRCode dll is not loaded or the QRCode struct not created. Execute _QRCode_StartUp() first. @extended contains 1 for the dll and 2 for the struct
; Author ........: Kanashius
; ==================================================================================================
Func _QRCode_GetQRCode($data, $iErrorLevel=$_QRCode_ECL_HIGH)
	If Not (IsString($data) or IsBinary($data)) Then return SetError(1, 1, -1)
	If Not ($iErrorLevel=$_QRCode_ECL_HIGH or $iErrorLevel=$_QRCode_ECL_LOW or $iErrorLevel=$_QRCode_ECL_MEDIUM or $iErrorLevel=$_QRCode_ECL_QUARTILE) Then return SetError(1, 2, -1)
	If $_QRCode_DLL=-1 Then SetError(8, 1, -1)
	If $_QRCode_tQRCode=-1 Then SetError(8, 2, -1)

	Local $ptrResult
	If IsString($data) Then
		$ptrResult = DllCall($_QRCode_DLL, "byte:cdecl", "EncodeText", "str", $data, "uint", StringLen($data), "byte", $iErrorLevel, "ptr", DllStructGetPtr($_QRCode_tQRCode))
		If @error Then return SetError(2, @error, -1)
	Else ; binary
		Local $iLength = StringLen($data)
		$ptrResult = DllCall($_QRCode_DLL, "byte:cdecl", "EncodeBinary", "str", $data, "uint", $iLength, "byte", $iErrorLevel, "ptr", DllStructGetPtr($_QRCode_tQRCode))
		If @error Then return SetError(2, @error, -1)
	EndIf

	Local $iError = DllStructGetData($_QRCode_tQRCode, "error")
	If $iError<>0 Then
		If $iError = 1 Then return SetError(4, 0, -1) ; text could not be read from input
		If $iError = 2 Then return SetError(5, 0, -1) ; segment to long
		If $iError = 3 Then
			Local $iDiff = DllStructGetData($_QRCode_tQRCode, "neededbits")-DllStructGetData($_QRCode_tQRCode, "possiblebits")
			return SetError(6, $iDiff, -1) ; data over capacity
		EndIf
		return SetError(7, $iError, -1) ; unknown error
	EndIf

	Local $iWidth = DllStructGetData($_QRCode_tQRCode, "width")
	Local $iHeight = DllStructGetData($_QRCode_tQRCode, "height")
	Local $tData = DllStructCreate("boolean["&$iWidth*$iHeight&"]", DllStructGetData($_QRCode_tQRCode, "data"))
	If @error Then return SetError(3, @error, -1)
	Local $arData[$iWidth][$iHeight]
	For $i=1 to $iWidth*$iHeight
		$arData[Int(($i-1)/$iWidth)][Mod($i-1, $iHeight)] = DllStructGetData($tData, 1, $i)
	Next
	return $arData
EndFunc